------------------------------------------------------------------------
-- Copyright Notice:                                                    
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory, 
-- which is operated for the National Science Foundation by the    
-- Association of Universities for Research in Astronomy under    
-- Cooperative Agreement No. AST-0132798.                                                                                                                           
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   SYSTEM TOP LEVEL  
-- MODULE       :   WISHBONE BUS SLAVE INTERFACE
-- VERSION      :   2.07 
------------------------------------------------------------------------
-- FILE NAME    :   WBSLAVEINTERFACE.VHD
-- AUTHOR       :   PETER MOORE
-- DATE         :   04/31/2011
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.02
------------------------------------------------------------------------
-- DESCRIPTION  :
-- Slave bus interface module connects a 'client module' to the standard 
-- Wishbone bus structure and supports normal 32-bit data reads and writes
-- within a 2^16 word address space. Write timing for normal transactions 
-- depends on client WrtAck signal from client interface logic for timing.
-- The read timing is set for 3 CLK_I cycle latency i.e. the slave interface
-- expects that the client data will be available 3 clock cycles after any
-- address change (this can be modified).
-- When no active read bus cycle is in progress, the slave interface
-- synchronously commutates the status_in data onto the DAT_O bus. This
-- System status data is latched into the INTERCON module and distributed 
-- to all slave interfaces as an out of band status word. The slave interface
-- latches this status word when ever the bus is idle.          
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- Bus Cycle Tag identifiers (TGC_I)
-- TGC_I + WE_I  
--   00 = Normal bus write cycle
--   01 = spare
--   10 = Start exposure cycle - produces a start exposure strobe and registers the start exposure vector.
--   11 = Local module reset - produces a reset strobe at the user interface port.
-- TGC_I + /WE_I
--   00 = normal bus read cycle
--   01 = spare
--   10 = spare
--   11 = spare
--
-- Bus Data Tag identifiers (TGD_I)  - Identifies the origin or destination of the bus transaction
-- TGD_I/O
-- FPDP bit 0
-- SYNC bit 1
-- UART bit 2
-- UART bit 3
--
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	  AUTHOR  VERSION  MODIFICATIONS
-- 06/03/2008  PCM   1.00   ORIGINAL VERSION
-- 10/02/2008  PCM   1.01   1st WORKING VERSION - BETA !
-- 06/01/2009  PCM   1.02   ADDED START EXPOSURE CAPABILITY
-- 03/19/2010   PCM   1.06  An attempt at reducing the write bus cycle count and making burst write available whilst solving the problem of double write strobes - VERSION NUMBER UPDATED TO BE COMMON WITH MASTER AND INTERCON
-- 04/31/2011    PCM   2.07  Split to create two versions: 1 with start vector port and one without start vector port. Mainly to reduce the number of false warning messages. 
------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.ALL;

entity WbSlaveInterface_Ver207a is
   port ( ADR_I       : in    std_logic_vector (15 downto 0);  -- Wishbone address input 
          CLK_I       : in    std_logic;                       --  Wishbone / System clock source
          CYC_I       : in    std_logic;                       -- Wishbone active cycle signal 
          DAT_I       : in    std_logic_vector (31 downto 0);  -- Wishbone data bus in
          RST_I       : in    std_logic;                       -- Wishbone / system reset signal in
          STB_I       : in    std_logic;                       --  Wishbone module select signal - Unique board / Module select strobe
          TGC_I       : in    std_logic_vector (1 downto 0);   -- Wishbone Cycle type identifier signals - Indicates what type of bus cycle this is 
          TGD_I       : in    std_logic_vector (3 downto 0);   -- Wishbone Data in type signal (input) - Indicates where the source Comms link for this bus transaction came from.
          WE_I        : in    std_logic;                       -- Wishbone Write Enable signal 
          ACK_O       : out   std_logic;                       --  Wishbone Write Cycle normal acknowledge signal
          DAT_O       : out   std_logic_vector (31 downto 0);  -- Wishbone data bus out - Read data back to the master 
          ERR_O       : out   std_logic;                       -- Wishbone Write Cycle Error acknowledge signal - Invalid address
          RTY_O       : out   std_logic;                       -- Wishbone Write Cycle retry acknowledge signal - Not used
          TGD_O       : out   std_logic_vector (3 downto 0);   -- Wishbone data out type signal  -  Indicates which Comms link this read data should go to
          DataIn      : in    std_logic_vector (31 downto 0);  -- 32-bit data input from Module for Wishbone read transactions  - data sent to master
          WrtAck      : in    std_logic;                       -- Write acknowledge from Modules for write transaction
          ErrorAck    : in    std_logic;                       -- Error marker from Module for write addressing error
          Busy        : in    std_logic;                       -- Busy / holdoff strobe from Modules during write / read transaction  
          SysClk      : out   std_logic;                       -- Module System clock
          Reset       : out   std_logic;                       -- Module master reset signal - Software or hardware generated
          DataOut     : out   std_logic_vector (31 downto 0);  -- 32-bit data to Module -  valid during write transactions
          Address     : out   std_logic_vector (15 downto 0);  -- 16-bit module address valid for write and read transactions 
          WriteEnable : out   std_logic;                       -- Write strobe to Module
          StatusOut   : out   std_logic_vector (31 downto 0);  -- System status received from the system interconnect for use by the module. 
          StatusIn    : in    std_logic_vector (7 downto 0);   -- Module status data to be transmitted to the system interconnect for distribution
          StartExpStb : out   std_logic;                       -- System wide start exposure strobe
          StartExpVector : out  std_logic_vector(15 downto 0)  -- System wide start exposure start vector
         );
end WbSlaveInterface_Ver207a;

architecture Behavioral of WbSlaveInterface_Ver207a is
   type RdStateType is (RdIdle, RdState1, RdState2, RdState3);
   signal RdState : RdStateType;
   signal ActiveCycle, RdAck : std_logic;
   signal StatDly : std_logic_vector(2 downto 0); -- Delay mechanism for status in enable
   signal StatusEnable : std_logic;
   signal NormalWrtCycle, NormalRdCycle, ResetWrtCycle : std_logic;
   signal LocalWriteStb, LocalReset, LocalBusy, RdBusy : std_logic;
   signal AddressReg : std_logic_vector(15 downto 0);
   signal DataInReg, StatusInReg  : std_logic_vector(31 downto 0);
   signal StartExpVectReg : std_logic_vector(15 downto 0);
   signal StartExpCycle : std_logic;
begin

------------------------
-- Static assignments --
------------------------
-- Slave Interface Input Signals
   SysClk <= CLK_I;                      -- System clock drives everything
   ActiveCycle <= STB_I and CYC_I;       -- Slave only responds when these two signals are true
   NormalWrtCycle <= not(TGC_I(0) or TGC_I(1) or LocalWriteStb) and WE_I and not RST_I;
   NormalRdCycle <= not(TGC_I(0) or TGC_I(1) or WE_I);
   ResetWrtCycle <= TGC_I(0) and TGC_I(1) and WE_I;
   StartExpCycle <= not TGC_I(0) and TGC_I(1) and WE_I;
   TGD_O <= TGD_I;
   
-- Slave Interface Output Signals
   WriteEnable <= LocalWriteStb and ActiveCycle;  --  and NormalWrtCycle;
   StartExpVector <= StartExpVectReg;
   Reset <= LocalReset;
   StatusOut <= StatusInReg;
   ACK_O <= not (ErrorAck or Busy) and (ActiveCycle and (WrtAck or RdAck or ResetWrtCycle));
   ERR_O <= not Busy and (ErrorAck and LocalWriteStb); 
   RTY_O <= ActiveCycle and Busy; 
   DataOut <= DataInReg;
   Address <= AddressReg;
   
-- Internal signals
   LocalBusy <= LocalWriteStb or NormalWrtCycle or RdBusy;
   
----------------------------------
-- Bus => interface Write logic --
----------------------------------

-- Data capture --
   WBWrt2SlvDataProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         StartExpStb <= '0';
         if (ActiveCycle = '1') then
            AddressReg <= ADR_I;        -- Latch the write address 
            if (NormalWrtCycle = '1') then
               DataInReg  <= DAT_I;        -- Latch the bus data onto our clients interface as output data.
            elsif (StartExpCycle = '1' and LocalBusy = '0') then
               StartExpVectReg <= ADR_I;
               StartExpStb <= '1';
            end if;
         elsif (StatusEnable = '1') then
            StatusInReg <= DAT_I;         -- Collect the global status on non-active-cycle clock edge  
         end if;
      end if;
   end process;
   
   StatusHoldOffProc : process(RST_I, CYC_I, CLK_I)
   begin
      if (RST_I = '1' or CYC_I = '1') then
         StatDly <= (others => '0');
      elsif (rising_edge(CLK_I)) then
         StatDly <= StatDly(1 downto 0) & '1';    -- Delay status capture for three clock cycles after any activity
      end if;
   end process;

   StatusEnable <= StatDly(2) and StatDly(1) and StatDly(0);
   
-- Local write strobe generator --
   WBWrt2SlvWrtStrobeProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            LocalWriteStb <= '0';
         elsif (LocalWriteStb = '1' and WrtAck = '1') then
            LocalWriteStb <= '0';         -- and cancel our data available strobe to the client.
         elsif (ActiveCycle = '1' and NormalWrtCycle = '1') then
            LocalWriteStb <= '1';         -- Detect a bus write and inform our client that data is available
         else
            LocalWriteStb <= '0';
         end if;
      end if;
   end process;

-- Local reset strobe generator --
   WBSlvResetProc : process (CLK_I)
   begin     
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            LocalReset <= '1';
         elsif (ActiveCycle = '1' and ResetWrtCycle = '1') then
            LocalReset <= '1';            -- Detect a local reset pulse writen by the bus.
         else
            LocalReset <= '0';
         end if;
      end if;
   end process;
   
---------------------------------
-- Bus => interface Read logic --
---------------------------------

-- read state machine - allows for logic delay through local logic --
   WBRd2SlvDataProc : process (CLK_I)
   begin
      if (rising_edge(CLK_I)) then
         if (RST_I = '1') then
            RdBusy <= '0';
            RdAck <= '0';
            DAT_O <= x"000000" & StatusIn;   -- Send 8-bits of local client status whenever not an active read cycle.
            RdState <= RdIdle;
         else
            case RdState is
               when RdIdle =>
                  if (ActiveCycle = '1' and NormalRdCycle = '1') then
                     RdBusy <= '1';
                     RdState <= RdState1;
                  else
                     RdBusy <= '0';
                     DAT_O <= x"000000" & StatusIn;   -- Send local status whenever not an active read cycle.
                     RdState <= RdIdle;
                  end if;
               when RdState1 =>                       -- Allow the address time to propgate to the internal logic.
                  RdState <= RdState2;
               when RdState2 =>
                  DAT_O <= DataIn;                    -- Latch the data from the internal logic 2 clock cycles after command.
                  RdAck <= '1';
                  RdState <= RdState3;
               when RdState3 =>                       -- Wait for the end of the bus cycle
                  if (ActiveCycle = '1') then
                     RdState <= RdState3;
                  else
                     RdAck <= '0';
                     DAT_O <= x"000000" & StatusIn;   -- Send local status whenever not an active read cycle.
                     RdState <= RdIdle;
                  end if;
            end case;
         end if;
      end if;
   end process;
   
end BEHAVIORAL;
